"""
This module contains compatibility helpers for Python 2/3 interoperability.

It mainly exists because their are certain incompatibilities in the Python
syntax that can only be solved by conditionally importing different functions.
"""
import sys
from lxml.etree import tostring

if sys.version_info[0] == 2:
    bytes_ = str
    str_ = unicode
    def tostring_(s):
        return tostring(s, encoding='utf-8').decode('utf-8')

elif sys.version_info[0] == 3:
    bytes_ = bytes
    str_ = str
    def tostring_(s):
        return tostring(s, encoding='utf-8')



def raise_with_traceback(exc_type, traceback, *args, **kwargs):
    """
    Raise a new exception of type `exc_type` with an existing `traceback`. All
    additional (keyword-)arguments are forwarded to `exc_type`
    """
    raise exc_type(*args, **kwargs), None, traceback



import errno
import os
import sys

import click

from r2k.cli import cli_utils, logger
from r2k.config import Config
from r2k.constants import Parser

from .prompts import Prompt


@click.command("init")
@cli_utils.config_path_option(exists=False)
@cli_utils.force_option("If set will override an existing configuration file (if it exists)")
def config_init(path: str, force: bool) -> None:
    """Initialize a new configuration files."""
    create_config_dir_if_not_exists(path)
    validate_config_overwrite(path, force)

    new_config = get_new_config()
    new_config.save(path)
    logger.info("Successfully set the configuration!")


def get_new_config() -> Config:
    """Ask the user to fill in all the necessary information, and return the resulting dict"""
    send_from = Prompt.get("send_from")
    password = Prompt.get("password")
    kindle_address = Prompt.get("kindle_address")
    parser = Prompt.get("parser")
    return Config(
        feeds={}, send_from=send_from, kindle_address=kindle_address, password=password, parser=Parser(parser)
    )


def validate_config_overwrite(path: str, force: bool) -> None:
    """Make sure the --force flag is passed where applicable"""
    if os.path.exists(path):
        if force:
            logger.warning(f"Overwriting existing configuration file in `{path}`...")
        else:
            logger.error(f"A file already exists in `{path}`.\nPass the --force flag to overwrite it.")
            sys.exit(1)


def create_config_dir_if_not_exists(path: str) -> None:
    """Create the config dir if it doesn't already exist"""
    dirname = os.path.dirname(path)
    try:
        os.makedirs(dirname)
    except OSError as exc:
        if exc.errno != errno.EEXIST:
            raise
        pass



from enum import Enum
from os.path import dirname, expanduser, join

CONFIG_ENV_VAR = "RSS_TO_KINDLE_CONFIG"

HTML_HEADERS = {
    "User-Agent": "Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/41.0.2228.0 Safari/537.3"
}

DEFAULT_APP_PATH = expanduser("~/.r2k")
DEFAULT_CONFIG_PATH = join(DEFAULT_APP_PATH, "config.yml")

PACKAGE_DIR = dirname(__file__)
TOP_LEVEL_DIR = dirname(PACKAGE_DIR)
TEMPLATES_DIR = join(TOP_LEVEL_DIR, "templates")

# Number of articles to put in a single EPUB eBook. Otherwise the email size might exceed GMAIL's 25MB limit
ARTICLE_EBOOK_LIMIT = 20


class Parser(Enum):
    """A convenience class to represent the available parsing options"""

    PUSH_TO_KINDLE = "pushtokindle"
    MERCURY = "mercury"
    READABILITY = "readability"

    __values__ = PUSH_TO_KINDLE, MERCURY, READABILITY



import sys
import urllib.parse

import click
import feedparser
import requests
from bs4 import BeautifulSoup
from pick import pick

from r2k.cli import cli_utils, logger
from r2k.config import config


@click.command("add")
@cli_utils.config_path_option()
@click.option("-t", "--title", type=str, required=True, help="The title of the feed")
@click.option("-u", "--url", type=str, required=True, help="The URL of the feed")
@cli_utils.force_option("If set will update existing feeds")
def feed_add(title: str, url: str, force: bool) -> None:
    """Add an RSS feed."""
    validate_existing_feeds(title, force)

    feeds = get_feeds_from_url(url)

    if not feeds:
        logger.error("Could not find an RSS feed")
        sys.exit(1)
    elif len(feeds) == 1:
        feed = feeds[0]
    else:
        feed, _ = pick(feeds, "Please choose the correct feed from this list:")

    config.feeds[title] = {"url": feed}
    config.save()
    logger.info("Successfully added the feed!")


def validate_existing_feeds(title: str, force: bool) -> None:
    """Error out if no force flag was passed and the feed already exists"""
    if title in config.feeds:
        if force:
            logger.confirm(f"Going to overwrite the following feed: {title}")
        else:
            logger.error(
                f"The following feed already exists: {title}\nPass the --force flag if you'd like to overwrite it"
            )
            sys.exit(1)


def get_feeds_from_url(url: str) -> list:
    """
    Try to parse the URL and find any RSS feeds in the webpage

    Adapted from: https://gist.github.com/alexmill/9bc634240531d81c3abe
    """
    logger.info(f"Attempting to find RSS feeds from {url}...")

    # If the URL itself is a proper RSS feed, just return it
    if is_rss_feed(url):
        logger.debug("URL is already a proper RSS feed")
        return [url]

    html = get_html(url)
    possible_feeds = get_feeds_from_links(html) + get_feeds_from_atags(url, html)

    return [url for url in set(possible_feeds) if is_rss_feed(url)]


def get_html(url: str) -> BeautifulSoup:
    """Parse the URL with bs4"""
    raw = requests.get(url).text
    return BeautifulSoup(raw, features="lxml")


def get_feeds_from_links(html: BeautifulSoup) -> list:
    """Try to find RSS feeds from link elements in the webpage"""
    possible_feeds = []

    feed_urls = html.findAll("link", rel="alternate")
    for f in feed_urls:
        t = f.get("type", None)
        if t:
            if "rss" in t or "xml" in t:
                href = f.get("href", None)
                if href:
                    possible_feeds.append(href)
    return possible_feeds


def get_feeds_from_atags(url: str, html: BeautifulSoup) -> list:
    """Try to find RSS feeds from a tags in the webpage"""
    possible_feeds = []

    parsed_url = urllib.parse.urlparse(url)
    if not (parsed_url.scheme and parsed_url.hostname):
        return []
    base = parsed_url.scheme + "://" + parsed_url.hostname
    atags = html.findAll("a")
    for a in atags:
        href = a.get("href", None)
        if href:
            if "xml" in href or "rss" in href or "feed" in href:
                possible_feeds.append(base + href)
    return possible_feeds


def is_rss_feed(url: str) -> bool:
    """Test whether the URL is a proper RSS feed"""
    f = feedparser.parse(url)
    return len(f.entries) > 0



import os
import sys
from typing import List

import arrow
import click

from r2k.cli import cli_utils, logger
from r2k.config import config
from r2k.constants import ARTICLE_EBOOK_LIMIT, Parser
from r2k.dates import get_pretty_date_str, now
from r2k.ebook.epub_builder import create_epub
from r2k.ebook.single_article import SingleArticle
from r2k.email_sender import send_epub, send_urls
from r2k.feeds import Article, Feed


@click.command("send")
@cli_utils.config_path_option()
@click.option(
    "-f",
    "--feed-title",
    type=str,
    required=False,
    help="Title of a (optional) feed. If not passed, updates for all feeds will be sent",
)
@click.option(
    "-u", "--url", type=str, required=False, help="URL of an article to send to the Kindle",
)
def kindle_send(feed_title: str, url: str) -> None:
    """Send updates from one or all feeds (or a single article)."""
    validate_parser()
    logger.info(f"[Parsing articles with the `{config.parser}` parser]\n")
    if feed_title:
        send_articles_for_feed(feed_title)
    elif url:
        send_article_from_url(url)
    else:
        logger.notice("Sending articles from all feeds...\n")
        for feed_title in config.feeds:
            send_articles_for_feed(feed_title)


def validate_parser() -> None:
    """Run various validations for the different parsers"""
    parser = Parser(config.parser)
    if parser == Parser.MERCURY:
        try:
            import docker  # noqa
        except ModuleNotFoundError:
            logger.error(
                "The `docker` module is not installed, but is required to use the `mercury` parser\n"
                "Consider either switching to a different parser (by running `r2k config set -k parser --force`)\n"
                "Or install the optional `docker` library by running `pip install 'r2k[docker]'`"
            )
            sys.exit(1)


def send_article_from_url(url: str) -> None:
    """Convert a single article using the parser, and send it to kindle as an Ebook"""
    logger.notice("Parsing article...")
    article = SingleArticle(url)
    send_updates([article], article.title)


def send_articles_for_feed(feed_title: str) -> None:
    """Find all the new/unread articles for a certain feed and send them to the user's kindle"""
    logger.notice(f"\nNow working on `{feed_title}`...")

    local_feed = get_local_feed(feed_title)
    unread_articles = get_unread_articles_for_feed(local_feed, feed_title)
    send_updates(unread_articles, feed_title)

    local_feed["updated"] = arrow.utcnow()
    config.save()


def get_unread_articles_for_feed(local_feed: dict, feed_title: str) -> List[Article]:
    """Find the all new articles for a certain feed"""
    rss_feed = Feed(local_feed["url"], feed_title)
    last_updated = local_feed.get("updated")
    return rss_feed.get_unread_articles(last_updated)


def get_local_feed(feed_title: str) -> dict:
    """Get the feed if it exists in the feeds dict, or exit with an error"""
    feed = config.feeds.get(feed_title)
    if not feed:
        logger.error(f"Tried to fetch articles from an unknown feed `{feed_title}`")
        sys.exit(1)
    return feed


def send_updates(unread_articles: List[Article], feed_title: str) -> None:
    """Iterate over `unread_articles`, and send each one to the kindle"""
    if unread_articles:
        if Parser(config.parser) == Parser.PUSH_TO_KINDLE:
            successful_count = send_urls([(article.title, article.link) for article in unread_articles])
        else:
            successful_count = send_epub_books(unread_articles, feed_title)

        if successful_count:
            logger.notice(f"Successfully sent {successful_count} articles from the `{feed_title}` feed!")
        else:
            logger.error(f"Failed to send any articles to `{feed_title}`. See errors above")
    else:
        logger.info(f"No new content for `{feed_title}`")


def send_epub_books(unread_articles: List[Article], feed_title: str) -> int:
    """
    Chunk the list of unread articles into chunks of max size ARTICLE_EBOOK_LIMIT

    This is in order to avoid creating too large of an EPUB and exceeding GMAIL's 25MB attachment size limit
    """
    successful_count = 0
    for i in range(0, len(unread_articles), ARTICLE_EBOOK_LIMIT):
        successful_count += send_epub_book(unread_articles[i : i + ARTICLE_EBOOK_LIMIT], feed_title)
    return successful_count


def send_epub_book(unread_articles: List[Article], feed_title: str) -> int:
    """Create an EPUB book from all the unread articles and send it via email"""
    date_range = get_unread_articles_date_range(unread_articles)
    title = f"{feed_title} [{date_range}]"
    epub_book = create_epub(unread_articles, title)
    try:
        success = send_epub(title, epub_book)
    finally:
        os.remove(epub_book)
    return len(unread_articles) if success else 0


def get_unread_articles_date_range(unread_articles: List[Article]) -> str:
    """Return a nicely formatted string with the range of dates for the article list"""
    if len(unread_articles) == 1:
        date = unread_articles[0].get_parsed_date()
        show_year = date.year != now().year
        return get_pretty_date_str(date, show_year=show_year)
    else:
        first_date = unread_articles[0].get_parsed_date()
        last_date = unread_articles[-1].get_parsed_date()
        show_year = not (first_date.year == last_date.year == now().year)
        first_date_str = get_pretty_date_str(first_date, show_year=show_year)
        last_date_str = get_pretty_date_str(last_date, show_year=show_year)
        return f"{first_date_str} — {last_date_str}"



from __future__ import annotations

from abc import ABC, abstractmethod
from types import TracebackType
from typing import Optional, Type

from bs4 import BeautifulSoup


class ParserBase(ABC):
    """
    Base class for parsers
    """

    @abstractmethod
    def __enter__(self) -> ParserBase:
        """Context manager __enter__"""
        pass

    @abstractmethod
    def __exit__(
        self, exc_type: Optional[Type[BaseException]], exc_val: Optional[BaseException], exc_tb: Optional[TracebackType]
    ) -> bool:
        """Context manager __exit__"""
        pass

    @abstractmethod
    def parse(self, url: str) -> dict:
        """Main function used to parse the article from a URL"""
        pass

    @staticmethod
    def fix_blockquotes(html: str) -> str:
        """Mobi doesn't seem to deal well with <p> tags inside <blockquote> tags. So we replace <p> with <div>"""
        soup = BeautifulSoup(html, "html.parser")
        for quote in soup.find_all("blockquote"):
            if not quote.p:
                continue
            quote.p.wrap(soup.new_tag("div"))  # Wrap all <p> elements with <div>
            quote.p.unwrap()  # Unwrap removes the element and replaces it with its content

        return soup.decode()



import re


# FIXME: use with caution, can leak memory
uids = {}
uids_document = None


def describe_node(node):
    global uids
    if node is None:
        return ""
    if not hasattr(node, "tag"):
        return "[%s]" % type(node)
    name = node.tag
    if node.get("id", ""):
        name += "#" + node.get("id")
    if node.get("class", "").strip():
        name += "." + ".".join(node.get("class").split())
    if name[:4] in ["div#", "div."]:
        name = name[3:]
    if name in ["tr", "td", "div", "p"]:
        uid = uids.get(node)
        if uid is None:
            uid = uids[node] = len(uids) + 1
        name += "{%02d}" % uid
    return name


def describe(node, depth=1):
    global uids, uids_document
    doc = node.getroottree().getroot()
    if doc != uids_document:
        uids = {}
        uids_document = doc

    # return repr(NodeRepr(node))
    parent = ""
    if depth and node.getparent() is not None:
        parent = describe(node.getparent(), depth=depth - 1) + ">"
    return parent + describe_node(node)


RE_COLLAPSE_WHITESPACES = re.compile(r"\s+", re.U)


def text_content(elem, length=40):
    content = RE_COLLAPSE_WHITESPACES.sub(" ", elem.text_content().replace("\r", ""))
    if len(content) < length:
        return content
    return content[:length] + "..."



import sys

import click

from r2k.cli import cli_utils, logger
from r2k.config import config as _config

from .prompts import Prompt


@click.command("set")
@click.option("-k", "--key", type=click.Choice(_config.fields()), required=True, help="The config key to set")
@cli_utils.config_path_option()
@cli_utils.force_option("If set will override an existing value (if it exists)")
def config_set(key: str, force: bool) -> None:
    """Set a value in the config."""
    if getattr(_config, key):
        if force:
            logger.warning("Overriding an existing value...")
        else:
            logger.error(f"A value already exists for `{key}`.\nPass the --force flag to overwrite it.")
            sys.exit(1)

    value = Prompt.get(key)

    setattr(_config, key, value)
    logger.info("Configuration successfully updated!")



#!/usr/bin/env python
from __future__ import print_function
import logging
import re
import sys

from lxml.etree import tounicode
from lxml.html import document_fromstring
from lxml.html import fragment_fromstring

from .cleaners import clean_attributes
from .cleaners import html_cleaner
from .htmls import build_doc
from .htmls import get_body
from .htmls import get_title
from .htmls import shorten_title
from .compat import str_, bytes_, tostring_
from .debug import describe, text_content


log = logging.getLogger("readability.readability")

REGEXES = {
    "unlikelyCandidatesRe": re.compile(
        r"combx|comment|community|disqus|extra|foot|header|menu|remark|rss|shoutbox|sidebar|sponsor|ad-break|agegate|pagination|pager|popup|tweet|twitter",
        re.I,
    ),
    "okMaybeItsACandidateRe": re.compile(r"and|article|body|column|main|shadow", re.I),
    "positiveRe": re.compile(
        r"article|body|content|entry|hentry|main|page|pagination|post|text|blog|story|comic|box",
        re.I,
    ),
    "negativeRe": re.compile(
        r"combx|comment|com-|contact|foot|footer|footnote|masthead|media|meta|outbrain|promo|related|scroll|shoutbox|sidebar|sponsor|shopping|tags|tool|widget|licenseText|bottom",
        re.I,
    ),
    "divToPElementsRe": re.compile(
        r"<(a|blockquote|dl|div|img|ol|p|pre|table|ul)", re.I
    ),
    #'replaceBrsRe': re.compile(r'(<br[^>]*>[ \n\r\t]*){2,}',re.I),
    #'replaceFontsRe': re.compile(r'<(\/?)font[^>]*>',re.I),
    #'trimRe': re.compile(r'^\s+|\s+$/'),
    #'normalizeRe': re.compile(r'\s{2,}/'),
    #'killBreaksRe': re.compile(r'(<br\s*\/?>(\s|&nbsp;?)*){1,}/'),
    "videoRe": re.compile(r"https?:\/\/(www\.)?(youtube|vimeo)\.com", re.I),
    # skipFootnoteLink:      /^\s*(\[?[a-z0-9]{1,2}\]?|^|edit|citation needed)\s*$/i,
}


class Unparseable(ValueError):
    pass


def to_int(x):
    if not x:
        return None
    x = x.strip()
    if x.endswith("px"):
        return int(x[:-2])
    if x.endswith("em"):
        return int(x[:-2]) * 12
    return int(x)


def clean(text):
    # Many spaces make the following regexes run forever
    text = re.sub(r"\s{255,}", " " * 255, text)
    text = re.sub(r"\s*\n\s*", "\n", text)
    text = re.sub(r"\t|[ \t]{2,}", " ", text)
    return text.strip()


def text_length(i):
    return len(clean(i.text_content() or ""))


def compile_pattern(elements):
    if not elements:
        return None
    elif isinstance(elements, re.Pattern):
        return elements
    elif isinstance(elements, (str_, bytes_)):
        if isinstance(elements, bytes_):
            elements = str_(elements, "utf-8")
        elements = elements.split(u",")
    if isinstance(elements, (list, tuple)):
        return re.compile(u"|".join([re.escape(x.strip()) for x in elements]), re.U)
    else:
        raise Exception("Unknown type for the pattern: {}".format(type(elements)))
        # assume string or string like object


class Document:
    """Class to build a etree document out of html."""

    def __init__(
        self,
        input,
        positive_keywords=None,
        negative_keywords=None,
        url=None,
        min_text_length=25,
        retry_length=250,
        xpath=False,
        handle_failures="discard",
    ):
        """Generate the document

        :param input: string of the html content.
        :param positive_keywords: regex, list or comma-separated string of patterns in classes and ids
        :param negative_keywords: regex, list or comma-separated string in classes and ids
        :param min_text_length: Tunable. Set to a higher value for more precise detection of longer texts.
        :param retry_length: Tunable. Set to a lower value for better detection of very small texts.
        :param xpath: If set to True, adds x="..." attribute to each HTML node,
        containing xpath path pointing to original document path (allows to
        reconstruct selected summary in original document).
        :param handle_failures: Parameter passed to `lxml` for handling failure during exception.
        Support options = ["discard", "ignore", None]

        Examples:
            positive_keywords=["news-item", "block"]
            positive_keywords=["news-item, block"]
            positive_keywords=re.compile("news|block")
            negative_keywords=["mysidebar", "related", "ads"]

        The Document class is not re-enterable.
        It is designed to create a new Document() for each HTML file to process it.

        API methods:
        .title() -- full title
        .short_title() -- cleaned up title
        .content() -- full content
        .summary() -- cleaned up content
        """
        self.input = input
        self.html = None
        self.encoding = None
        self.positive_keywords = compile_pattern(positive_keywords)
        self.negative_keywords = compile_pattern(negative_keywords)
        self.url = url
        self.min_text_length = min_text_length
        self.retry_length = retry_length
        self.xpath = xpath
        self.handle_failures = handle_failures

    def _html(self, force=False):
        if force or self.html is None:
            self.html = self._parse(self.input)
            if self.xpath:
                root = self.html.getroottree()
                for i in self.html.getiterator():
                    # print root.getpath(i)
                    i.attrib["x"] = root.getpath(i)
        return self.html

    def _parse(self, input):
        doc, self.encoding = build_doc(input)
        doc = html_cleaner.clean_html(doc)
        base_href = self.url
        if base_href:
            # trying to guard against bad links like <a href="http://[http://...">
            try:
                # such support is added in lxml 3.3.0
                doc.make_links_absolute(
                    base_href,
                    resolve_base_href=True,
                    handle_failures=self.handle_failures,
                )
            except TypeError:  # make_links_absolute() got an unexpected keyword argument 'handle_failures'
                # then we have lxml < 3.3.0
                # please upgrade to lxml >= 3.3.0 if you're failing here!
                doc.make_links_absolute(
                    base_href,
                    resolve_base_href=True,
                    handle_failures=self.handle_failures,
                )
        else:
            doc.resolve_base_href(handle_failures=self.handle_failures)
        return doc

    def content(self):
        """Returns document body"""
        return get_body(self._html(True))

    def title(self):
        """Returns document title"""
        return get_title(self._html(True))

    def short_title(self):
        """Returns cleaned up document title"""
        return shorten_title(self._html(True))

    def get_clean_html(self):
        """
        An internal method, which can be overridden in subclasses, for example,
        to disable or to improve DOM-to-text conversion in .summary() method
        """
        return clean_attributes(tounicode(self.html, method="html"))

    def summary(self, html_partial=False):
        """
        Given a HTML file, extracts the text of the article.

        :param html_partial: return only the div of the document, don't wrap
                             in html and body tags.

        Warning: It mutates internal DOM representation of the HTML document,
        so it is better to call other API methods before this one.
        """
        try:
            ruthless = True
            while True:
                self._html(True)
                for i in self.tags(self.html, "script", "style"):
                    i.drop_tree()
                for i in self.tags(self.html, "body"):
                    i.set("id", "readabilityBody")
                if ruthless:
                    self.remove_unlikely_candidates()
                self.transform_misused_divs_into_paragraphs()
                candidates = self.score_paragraphs()

                best_candidate = self.select_best_candidate(candidates)

                if best_candidate:
                    article = self.get_article(
                        candidates, best_candidate, html_partial=html_partial
                    )
                else:
                    if ruthless:
                        log.info("ruthless removal did not work. ")
                        ruthless = False
                        log.debug(
                            (
                                "ended up stripping too much - "
                                "going for a safer _parse"
                            )
                        )
                        # try again
                        continue
                    else:
                        log.debug(
                            (
                                "Ruthless and lenient parsing did not work. "
                                "Returning raw html"
                            )
                        )
                        article = self.html.find("body")
                        if article is None:
                            article = self.html
                cleaned_article = self.sanitize(article, candidates)

                article_length = len(cleaned_article or "")
                retry_length = self.retry_length
                of_acceptable_length = article_length >= retry_length
                if ruthless and not of_acceptable_length:
                    ruthless = False
                    # Loop through and try again.
                    continue
                else:
                    return cleaned_article
        except Exception as e:
            log.exception("error getting summary: ")
            if sys.version_info[0] == 2:
                from .compat.two import raise_with_traceback
            else:
                from .compat.three import raise_with_traceback
            raise_with_traceback(Unparseable, sys.exc_info()[2], str_(e))

    def get_article(self, candidates, best_candidate, html_partial=False):
        # Now that we have the top candidate, look through its siblings for
        # content that might also be related.
        # Things like preambles, content split by ads that we removed, etc.
        sibling_score_threshold = max([10, best_candidate["content_score"] * 0.2])
        # create a new html document with a html->body->div
        if html_partial:
            output = fragment_fromstring("<div/>")
        else:
            output = document_fromstring("<div/>")
        best_elem = best_candidate["elem"]
        parent = best_elem.getparent()
        siblings = parent.getchildren() if parent is not None else [best_elem]
        for sibling in siblings:
            # in lxml there no concept of simple text
            # if isinstance(sibling, NavigableString): continue
            append = False
            if sibling is best_elem:
                append = True
            sibling_key = sibling  # HashableElement(sibling)
            if (
                sibling_key in candidates
                and candidates[sibling_key]["content_score"] >= sibling_score_threshold
            ):
                append = True

            if sibling.tag == "p":
                link_density = self.get_link_density(sibling)
                node_content = sibling.text or ""
                node_length = len(node_content)

                if node_length > 80 and link_density < 0.25:
                    append = True
                elif (
                    node_length <= 80
                    and link_density == 0
                    and re.search(r"\.( |$)", node_content)
                ):
                    append = True

            if append:
                # We don't want to append directly to output, but the div
                # in html->body->div
                if html_partial:
                    output.append(sibling)
                else:
                    output.getchildren()[0].getchildren()[0].append(sibling)
        # if output is not None:
        #    output.append(best_elem)
        return output

    def select_best_candidate(self, candidates):
        if not candidates:
            return None

        sorted_candidates = sorted(
            candidates.values(), key=lambda x: x["content_score"], reverse=True
        )
        for candidate in sorted_candidates[:5]:
            elem = candidate["elem"]
            log.debug("Top 5 : %6.3f %s" % (candidate["content_score"], describe(elem)))

        best_candidate = sorted_candidates[0]
        return best_candidate

    def get_link_density(self, elem):
        link_length = 0
        for i in elem.findall(".//a"):
            link_length += text_length(i)
        # if len(elem.findall(".//div") or elem.findall(".//p")):
        #    link_length = link_length
        total_length = text_length(elem)
        return float(link_length) / max(total_length, 1)

    def score_paragraphs(self):
        MIN_LEN = self.min_text_length
        candidates = {}
        ordered = []
        for elem in self.tags(self._html(), "p", "pre", "td"):
            parent_node = elem.getparent()
            if parent_node is None:
                continue
            grand_parent_node = parent_node.getparent()

            inner_text = clean(elem.text_content() or "")
            inner_text_len = len(inner_text)

            # If this paragraph is less than 25 characters
            # don't even count it.
            if inner_text_len < MIN_LEN:
                continue

            if parent_node not in candidates:
                candidates[parent_node] = self.score_node(parent_node)
                ordered.append(parent_node)

            if grand_parent_node is not None and grand_parent_node not in candidates:
                candidates[grand_parent_node] = self.score_node(grand_parent_node)
                ordered.append(grand_parent_node)

            content_score = 1
            content_score += len(inner_text.split(","))
            content_score += min((inner_text_len / 100), 3)
            # if elem not in candidates:
            #    candidates[elem] = self.score_node(elem)

            # WTF? candidates[elem]['content_score'] += content_score
            candidates[parent_node]["content_score"] += content_score
            if grand_parent_node is not None:
                candidates[grand_parent_node]["content_score"] += content_score / 2.0

        # Scale the final candidates score based on link density. Good content
        # should have a relatively small link density (5% or less) and be
        # mostly unaffected by this operation.
        for elem in ordered:
            candidate = candidates[elem]
            ld = self.get_link_density(elem)
            score = candidate["content_score"]
            log.debug(
                "Branch %6.3f %s link density %.3f -> %6.3f"
                % (score, describe(elem), ld, score * (1 - ld))
            )
            candidate["content_score"] *= 1 - ld

        return candidates

    def class_weight(self, e):
        weight = 0
        for feature in [e.get("class", None), e.get("id", None)]:
            if feature:
                if REGEXES["negativeRe"].search(feature):
                    weight -= 25

                if REGEXES["positiveRe"].search(feature):
                    weight += 25

                if self.positive_keywords and self.positive_keywords.search(feature):
                    weight += 25

                if self.negative_keywords and self.negative_keywords.search(feature):
                    weight -= 25

        if self.positive_keywords and self.positive_keywords.match("tag-" + e.tag):
            weight += 25

        if self.negative_keywords and self.negative_keywords.match("tag-" + e.tag):
            weight -= 25

        return weight

    def score_node(self, elem):
        content_score = self.class_weight(elem)
        name = elem.tag.lower()
        if name in ["div", "article"]:
            content_score += 5
        elif name in ["pre", "td", "blockquote"]:
            content_score += 3
        elif name in ["address", "ol", "ul", "dl", "dd", "dt", "li", "form", "aside"]:
            content_score -= 3
        elif name in [
            "h1",
            "h2",
            "h3",
            "h4",
            "h5",
            "h6",
            "th",
            "header",
            "footer",
            "nav",
        ]:
            content_score -= 5
        return {"content_score": content_score, "elem": elem}

    def remove_unlikely_candidates(self):
        for elem in self.html.findall(".//*"):
            s = "%s %s" % (elem.get("class", ""), elem.get("id", ""))
            if len(s) < 2:
                continue
            if (
                REGEXES["unlikelyCandidatesRe"].search(s)
                and (not REGEXES["okMaybeItsACandidateRe"].search(s))
                and elem.tag not in ["html", "body"]
            ):
                log.debug("Removing unlikely candidate - %s" % describe(elem))
                elem.drop_tree()

    def transform_misused_divs_into_paragraphs(self):
        for elem in self.tags(self.html, "div"):
            # transform <div>s that do not contain other block elements into
            # <p>s
            # FIXME: The current implementation ignores all descendants that
            # are not direct children of elem
            # This results in incorrect results in case there is an <img>
            # buried within an <a> for example
            if not REGEXES["divToPElementsRe"].search(
                str_(b"".join(map(tostring_, list(elem))))
            ):
                # log.debug("Altering %s to p" % (describe(elem)))
                elem.tag = "p"
                # print "Fixed element "+describe(elem)

        for elem in self.tags(self.html, "div"):
            if elem.text and elem.text.strip():
                p = fragment_fromstring("<p/>")
                p.text = elem.text
                elem.text = None
                elem.insert(0, p)
                # print "Appended "+tounicode(p)+" to "+describe(elem)

            for pos, child in reversed(list(enumerate(elem))):
                if child.tail and child.tail.strip():
                    p = fragment_fromstring("<p/>")
                    p.text = child.tail
                    child.tail = None
                    elem.insert(pos + 1, p)
                    # print "Inserted "+tounicode(p)+" to "+describe(elem)
                if child.tag == "br":
                    # print 'Dropped <br> at '+describe(elem)
                    child.drop_tree()

    def tags(self, node, *tag_names):
        for tag_name in tag_names:
            for e in node.findall(".//%s" % tag_name):
                yield e

    def reverse_tags(self, node, *tag_names):
        for tag_name in tag_names:
            for e in reversed(node.findall(".//%s" % tag_name)):
                yield e

    def sanitize(self, node, candidates):
        MIN_LEN = self.min_text_length
        for header in self.tags(node, "h1", "h2", "h3", "h4", "h5", "h6"):
            if self.class_weight(header) < 0 or self.get_link_density(header) > 0.33:
                header.drop_tree()

        for elem in self.tags(node, "form", "textarea"):
            elem.drop_tree()

        for elem in self.tags(node, "iframe"):
            if "src" in elem.attrib and REGEXES["videoRe"].search(elem.attrib["src"]):
                elem.text = "VIDEO"  # ADD content to iframe text node to force <iframe></iframe> proper output
            else:
                elem.drop_tree()

        allowed = {}
        # Conditionally clean <table>s, <ul>s, and <div>s
        for el in self.reverse_tags(
            node, "table", "ul", "div", "aside", "header", "footer", "section"
        ):
            if el in allowed:
                continue
            weight = self.class_weight(el)
            if el in candidates:
                content_score = candidates[el]["content_score"]
                # print '!',el, '-> %6.3f' % content_score
            else:
                content_score = 0
            tag = el.tag

            if weight + content_score < 0:
                log.debug(
                    "Removed %s with score %6.3f and weight %-3s"
                    % (describe(el), content_score, weight,)
                )
                el.drop_tree()
            elif el.text_content().count(",") < 10:
                counts = {}
                for kind in ["p", "img", "li", "a", "embed", "input"]:
                    counts[kind] = len(el.findall(".//%s" % kind))
                counts["li"] -= 100
                counts["input"] -= len(el.findall('.//input[@type="hidden"]'))

                # Count the text length excluding any surrounding whitespace
                content_length = text_length(el)
                link_density = self.get_link_density(el)
                parent_node = el.getparent()
                if parent_node is not None:
                    if parent_node in candidates:
                        content_score = candidates[parent_node]["content_score"]
                    else:
                        content_score = 0
                # if parent_node is not None:
                # pweight = self.class_weight(parent_node) + content_score
                # pname = describe(parent_node)
                # else:
                # pweight = 0
                # pname = "no parent"
                to_remove = False
                reason = ""

                # if el.tag == 'div' and counts["img"] >= 1:
                #    continue
                if counts["p"] and counts["img"] > 1 + counts["p"] * 1.3:
                    reason = "too many images (%s)" % counts["img"]
                    to_remove = True
                elif counts["li"] > counts["p"] and tag not in ("ol", "ul"):
                    reason = "more <li>s than <p>s"
                    to_remove = True
                elif counts["input"] > (counts["p"] / 3):
                    reason = "less than 3x <p>s than <input>s"
                    to_remove = True
                elif content_length < MIN_LEN and counts["img"] == 0:
                    reason = (
                        "too short content length %s without a single image"
                        % content_length
                    )
                    to_remove = True
                elif content_length < MIN_LEN and counts["img"] > 2:
                    reason = (
                        "too short content length %s and too many images"
                        % content_length
                    )
                    to_remove = True
                elif weight < 25 and link_density > 0.2:
                    reason = "too many links %.3f for its weight %s" % (
                        link_density,
                        weight,
                    )
                    to_remove = True
                elif weight >= 25 and link_density > 0.5:
                    reason = "too many links %.3f for its weight %s" % (
                        link_density,
                        weight,
                    )
                    to_remove = True
                elif (counts["embed"] == 1 and content_length < 75) or counts[
                    "embed"
                ] > 1:
                    reason = (
                        "<embed>s with too short content length, or too many <embed>s"
                    )
                    to_remove = True
                elif not content_length:
                    reason = "no content"
                    to_remove = True
                if el.tag == 'div' and counts['img'] >= 1 and to_remove:
                    if el.findall('.//img'):
                        to_remove = False
                        log.debug("Allowing %s" %el.text_content())
                        for desnode in self.tags(el, "table", "ul", "div"):
                            allowed[desnode] = True

                    # find x non empty preceding and succeeding siblings
                    i, j = 0, 0
                    x = 1
                    siblings = []
                    for sib in el.itersiblings():
                        # log.debug(sib.text_content())
                        sib_content_length = text_length(sib)
                        if sib_content_length:
                            i = +1
                            siblings.append(sib_content_length)
                            if i == x:
                                break
                    for sib in el.itersiblings(preceding=True):
                        # log.debug(sib.text_content())
                        sib_content_length = text_length(sib)
                        if sib_content_length:
                            j = +1
                            siblings.append(sib_content_length)
                            if j == x:
                                break
                    # log.debug(str_(siblings))
                    if siblings and sum(siblings) > 1000:
                        to_remove = False
                        log.debug("Allowing %s" % describe(el))
                        for desnode in self.tags(el, "table", "ul", "div", "section"):
                            allowed[desnode] = True

                if to_remove:
                    log.debug(
                        "Removed %6.3f %s with weight %s cause it has %s."
                        % (content_score, describe(el), weight, reason)
                    )
                    # print tounicode(el)
                    # log.debug("pname %s pweight %.3f" %(pname, pweight))
                    el.drop_tree()
                else:
                    log.debug(
                        "Not removing %s of length %s: %s"
                        % (describe(el), content_length, text_content(el))
                    )

        self.html = node
        return self.get_clean_html()


def main():
    VERBOSITY = {1: logging.WARNING, 2: logging.INFO, 3: logging.DEBUG}

    from optparse import OptionParser

    parser = OptionParser(usage="%prog: [options] [file]")
    parser.add_option("-v", "--verbose", action="count", default=0)
    parser.add_option(
        "-b", "--browser", default=None, action="store_true", help="open in browser"
    )
    parser.add_option(
        "-l", "--log", default=None, help="save logs into file (appended)"
    )
    parser.add_option(
        "-u", "--url", default=None, help="use URL instead of a local file"
    )
    parser.add_option("-x", "--xpath", default=None, help="add original xpath")
    parser.add_option(
        "-p",
        "--positive-keywords",
        default=None,
        help="positive keywords (comma-separated)",
        action="store",
    )
    parser.add_option(
        "-n",
        "--negative-keywords",
        default=None,
        help="negative keywords (comma-separated)",
        action="store",
    )
    (options, args) = parser.parse_args()

    if options.verbose:
        logging.basicConfig(
            level=VERBOSITY[options.verbose],
            filename=options.log,
            format="%(asctime)s: %(levelname)s: %(message)s (at %(filename)s: %(lineno)d)",
        )

    if not (len(args) == 1 or options.url):
        parser.print_help()
        sys.exit(1)

    file = None
    if options.url:
        headers = {"User-Agent": "Mozilla/5.0"}
        if sys.version_info[0] == 3:
            import urllib.request, urllib.parse, urllib.error

            request = urllib.request.Request(options.url, None, headers)
            file = urllib.request.urlopen(request)
        else:
            import urllib2

            request = urllib2.Request(options.url, None, headers)
            file = urllib2.urlopen(request)
    else:
        file = open(args[0], "rt")
    try:
        doc = Document(
            file.read(),
            url=options.url,
            positive_keywords=options.positive_keywords,
            negative_keywords=options.negative_keywords,
        )
        if options.browser:
            from .browser import open_in_browser

            result = "<h2>" + doc.short_title() + "</h2><br/>" + doc.summary()
            open_in_browser(result)
        else:
            enc = (
                sys.__stdout__.encoding or "utf-8"
            )  # XXX: this hack could not always work, better to set PYTHONIOENCODING
            result = "Title:" + doc.short_title() + "\n" + doc.summary()
            if sys.version_info[0] == 3:
                print(result)
            else:
                print(result.encode(enc, "replace"))
    finally:
        file.close()


if __name__ == "__main__":
    main()



import re
import chardet
import sys


RE_CHARSET = re.compile(br'<meta.*?charset=["\']*(.+?)["\'>]', flags=re.I)
RE_PRAGMA = re.compile(br'<meta.*?content=["\']*;?charset=(.+?)["\'>]', flags=re.I)
RE_XML = re.compile(br'^<\?xml.*?encoding=["\']*(.+?)["\'>]')

CHARSETS = {
    "big5": "big5hkscs",
    "gb2312": "gb18030",
    "ascii": "utf-8",
    "maccyrillic": "cp1251",
    "win1251": "cp1251",
    "win-1251": "cp1251",
    "windows-1251": "cp1251",
}


def fix_charset(encoding):
    """Overrides encoding when charset declaration
       or charset determination is a subset of a larger
       charset.  Created because of issues with Chinese websites"""
    encoding = encoding.lower()
    return CHARSETS.get(encoding, encoding)


def get_encoding(page):
    # Regex for XML and HTML Meta charset declaration
    declared_encodings = (
        RE_CHARSET.findall(page) + RE_PRAGMA.findall(page) + RE_XML.findall(page)
    )

    # Try any declared encodings
    for declared_encoding in declared_encodings:
        try:
            if sys.version_info[0] == 3:
                # declared_encoding will actually be bytes but .decode() only
                # accepts `str` type. Decode blindly with ascii because no one should
                # ever use non-ascii characters in the name of an encoding.
                declared_encoding = declared_encoding.decode("ascii", "replace")

            encoding = fix_charset(declared_encoding)

            # Now let's decode the page
            page.decode(encoding)
            # It worked!
            return encoding
        except (UnicodeDecodeError, LookupError):
            pass

    # Fallback to chardet if declared encodings fail
    # Remove all HTML tags, and leave only text for chardet
    text = re.sub(br"(\s*</?[^>]*>)+\s*", b" ", page).strip()
    enc = "utf-8"
    if len(text) < 10:
        return enc  # can't guess
    res = chardet.detect(text)
    enc = res["encoding"] or "utf-8"
    # print '->', enc, "%.2f" % res['confidence']
    enc = fix_charset(enc)
    return enc



# strip out a set of nuisance html attributes that can mess up rendering in RSS feeds
import re
from lxml.html.clean import Cleaner

bad_attrs = ["width", "height", "style", "[-a-z]*color", "background[-a-z]*", "on*"]
single_quoted = "'[^']+'"
double_quoted = '"[^"]+"'
non_space = "[^ \"'>]+"
htmlstrip = re.compile(
    "<"  # open
    "([^>]+) "  # prefix
    "(?:%s) *" % ("|".join(bad_attrs),)
    + "= *(?:%s|%s|%s)"  # undesirable attributes
    % (non_space, single_quoted, double_quoted)
    + "([^>]*)"  # value  # postfix
    ">",  # end
    re.I,
)


def clean_attributes(html):
    while htmlstrip.search(html):
        html = htmlstrip.sub("<\\1\\2>", html)
    return html


def normalize_spaces(s):
    if not s:
        return ""
    """replace any sequence of whitespace
    characters with a single space"""
    return " ".join(s.split())


html_cleaner = Cleaner(
    scripts=True,
    javascript=True,
    comments=True,
    style=True,
    links=True,
    meta=False,
    add_nofollow=False,
    page_structure=False,
    processing_instructions=True,
    embedded=False,
    frames=False,
    forms=False,
    annoying_tags=False,
    remove_tags=None,
    remove_unknown_tags=False,
    safe_attrs_only=False,
)



import click

from r2k.cli.config import config
from r2k.cli.feed import feed
from r2k.cli.kindle import kindle


@click.group()
@click.option("-v", "--verbose", is_flag=True, default=False, help="If set, will print DEBUG messages")
@click.option("--no-ansi", is_flag=True, default=False, help="Disable ANSI output")
@click.pass_context
def main(ctx: click.Context, verbose: bool, no_ansi: bool) -> None:
    """A tool to send items from RSS feeds to Kindle"""
    setattr(ctx, "verbose", verbose)
    setattr(ctx, "no_ansi", no_ansi)


main.add_command(config)
main.add_command(feed)
main.add_command(kindle)


if __name__ == "__main__":
    main()



"""Adapted from SO: https://stackoverflow.com/a/48946422/978089"""
import re
import unicodedata
from typing import Dict, List, Tuple, Union

_common_char_maps: Dict[str, str] = {}


def _unicode_character_name(char: str) -> Union[str, None]:
    """Get the full unicode char name"""
    try:
        return unicodedata.name(char)
    except ValueError:
        return None


def _get_all_unicode_characters() -> List[Tuple[str, str]]:
    """Generate all Unicode characters with their names"""
    all_unicode_characters = []
    for n in range(0, 0x10FFFF):  # Unicode planes 0-16
        char = chr(n)  # Python 3
        name = _unicode_character_name(char)
        if name:
            all_unicode_characters.append((char, name))
    return all_unicode_characters


def _get_common_char_map() -> Dict[str, str]:
    """Create a mapping from common unicode characters to their ASCII counterparts"""
    global _common_char_maps
    if not _common_char_maps:
        all_unicode_characters = _get_all_unicode_characters()
        for char, name in all_unicode_characters:
            if "DOUBLE QUOTATION MARK" in name:
                _common_char_maps[char] = '"'
            elif "SINGLE QUOTATION MARK" in name:
                _common_char_maps[char] = "'"
            elif "DASH" in name and "DASHED" not in name:
                _common_char_maps[char] = "-"
    return _common_char_maps


def strip_common_unicode_chars(string: str) -> str:
    """Replace common unicode characters with comparable ASCII values"""
    char_map = _get_common_char_map()
    for char, replacement in char_map.items():
        if char in string:
            string = string.replace(char, replacement)
    return string


def normalize_str(string: str) -> str:
    """
    Strip any non alpha-numeric characters from a string and replace them with underscores
    """
    return re.sub(r"\W+", "_", string)



from typing import Optional
from urllib.parse import unquote, urljoin, urlparse
from uuid import uuid4

import requests
from bs4.element import Tag

from r2k.constants import HTML_HEADERS


def download_image(url: str, path: str) -> None:
    """
    Download an image from URL to path
    """
    # download the body of response by chunk, not immediately
    response = requests.get(url, headers=HTML_HEADERS, stream=True)
    with open(path, "wb") as f:
        for data in response.iter_content(1024):
            # write data read to the file
            f.write(data)


def get_image_filename(url: str) -> str:
    """
    Get image name from its URL with a unique prefix to avoid collisions
    """
    image_basename = unquote(url).split("/")[-1]
    image_basename = image_basename[-150:]
    prefix = str(uuid4())[:8]
    # We start with img because XML IDs cannot start with numbers
    return f"img-{prefix}-{image_basename}"


def is_valid_img_url(url: str) -> bool:
    """
    Checks whether `url` is a valid URL.
    """
    parsed = urlparse(url)
    return bool(parsed.netloc) and bool(parsed.scheme)


def get_img_url(url: str, img: Tag) -> Optional[str]:
    """
    Get an absolute img URL from an img HTML element (with some checks and verifications first)
    """
    img_url = img.attrs.get("src")
    if not img_url:
        # if img does not contain src attribute, just skip
        return None

    # In case img_url is relative, append it to the root url (urljoin will return img_url if it's an absolute URL)
    img_url = urljoin(url, img_url)

    # remove URLs like '/hsts-pixel.gif?c=3.2.5'
    try:
        pos = img_url.index("?")
        img_url = img_url[:pos]
    except ValueError:
        pass

    if is_valid_img_url(img_url):
        return img_url
    return None


def get_img_extension(path: str) -> str:
    """
    Parse the path and return the image's extension (replacing jpg to jpeg with accordance with EPUB spec)
    """
    path_list = path.rsplit(".", 1)
    if len(path_list) == 2:
        ext = path_list[1]
    else:
        ext = "jpeg"
    if ext == "jpg":
        ext = "jpeg"
    return ext



def open_in_browser(html):
    """
    Open the HTML document in a web browser, saving it to a temporary
    file to open it.  Note that this does not delete the file after
    use.  This is mainly meant for debugging.
    """
    import os
    import webbrowser
    import tempfile

    handle, fn = tempfile.mkstemp(suffix=".html")
    f = os.fdopen(handle, "wb")
    try:
        f.write(b"<meta charset='UTF-8' />")
        f.write(html.encode("utf-8"))
    finally:
        # we leak the file itself here, but we should at least close it
        f.close()
    url = "file://" + fn.replace(os.path.sep, "/")
    webbrowser.open(url)
    return url



import click
import orjson as json
import yaml

from r2k.cli import cli_utils, logger
from r2k.config import config as _config


@click.command("show")
@cli_utils.config_path_option()
@click.option(
    "-j",
    "--json",
    "is_json",
    is_flag=True,
    help="When passed the output will be in JSON format (e.g. for use with jq).\n"
    "Use with the --no-ansi flag for best results",
)
def config_show(is_json: bool) -> None:
    """Show all the available configuration."""
    result = _config.as_dict()
    if "password" in result:
        result["password"] = "XXXXXXX"
    if is_json:
        logger.info(json.dumps(result))
    else:
        logger.info(yaml.safe_dump(result))



import requests
from bs4 import BeautifulSoup

from r2k.feeds import Article


class SingleArticle(Article):
    """Represents a single (i.e. not part of a feed) article for sending it to kindle"""

    def __init__(self, url: str) -> None:
        """Constructor"""
        super().__init__({"link": url})
        reqs = requests.get(self.link)
        self._soup = BeautifulSoup(reqs.text, "html.parser")

        self.set_title()
        self.set_author()

    def set_title(self) -> None:
        """Set the article's title"""
        titles = self._soup.find_all("title")
        if titles:
            self.title = titles[0].get_text().strip()
        else:
            self.title = self.link.strip()

    def set_author(self) -> None:
        """Set the article's author"""
        metas = self._soup.find_all("meta")
        for meta in metas:
            if meta.attrs.get("name") == "author":
                self.author = meta.attrs.get("content").strip()



from os import listdir, makedirs, walk
from os.path import join
from shutil import copyfile, rmtree
from string import Template
from tempfile import mkdtemp
from typing import Any, List, Optional, Type
from uuid import uuid4
from zipfile import ZIP_DEFLATED, ZIP_STORED, ZipFile

import arrow
from bs4 import BeautifulSoup

from r2k.cli import logger
from r2k.config import config
from r2k.constants import TEMPLATES_DIR, Parser
from r2k.feeds import Article
from r2k.unicode import normalize_str, strip_common_unicode_chars

from . import images
from .base_parser import ParserBase

META_INF = "META-INF"
OEBPS = "OEBPS"
MIMETYPE = "mimetype"
IMAGES = "images"
CONTENT = "content"
_R2K = "_r2k"

# The number of elements in toc.ncx that appear before the articles (the initial 1 is because toc starts from 1, not 0)
NAVPOINT_OFFSET = 1 + 2

EPUB_DIR = join(TEMPLATES_DIR, "epub")


def create_epub(raw_articles: List[Article], title: str) -> str:
    """
    Create an EPUB book from multiple articles.

    :returns temp path to created ebook
    """
    epub_path = mkdtemp()
    logger.debug(f"Creating epub folder in {epub_path}")

    articles = [EPUBArticle(raw_article, epub_path) for raw_article in raw_articles]
    book = EPUB(articles, title, epub_path)
    return book.build()


class EPUBArticle:
    """
    Represents a single article, with rendering and parsing options for transforming it into EPUB content
    """

    def __init__(self, raw_article: Article, epub_path: str):
        """
        Constructor
        """
        self.url = raw_article.link
        self.title = raw_article.title
        self.id = normalize_str(self.title)
        self.author = raw_article.get("author", "")
        self.date = raw_article.get_str_date()
        self.images_path = join(epub_path, OEBPS, IMAGES)

        self.content: Optional[str] = None
        self.images: List[str] = []

    def parse(self, parser: ParserBase) -> bool:
        """
        Prepare the content of the article for EPUB

        Performs these tasks:
            1. Parse the article with the Mercury parser
            2. Download all the images mentioned in the article (EPUB format only supports embedded local images)
            3. Replace all the <img "src"> tags with the local paths of the downloaded images
        :return: True iff the parsing succeeded
        """
        logger.info(f"Parsing `{self.title}`...")
        parsed_article = parser.parse(self.url)
        raw_content = parsed_article.get("content")
        if not raw_content:
            return False

        # When some blogs (like xkcd.com) are parsed,
        # their content doesn't include the actual image, so we're adding it here
        if lead_image_url := parsed_article.get("lead_image_url"):
            raw_content = f'<img src="{lead_image_url}"/>\n{raw_content}'
        clean_html = strip_common_unicode_chars(raw_content)
        self.content = self.parse_images(clean_html)
        return True

    def parse_images(self, raw_content: str) -> str:
        """
        Parse and download images in the article

        Go over the content of the article and:
            1. Find all the `img` tags in the HTML
            2. Download all the images to the `images` folder in the EPUB dir
            3. Set the relative paths to those images in the HTML content
            4. Update the `content` attribute with the new HTML content
        """
        soup = BeautifulSoup(raw_content, "html.parser")
        logger.debug("Looking for images...")
        for img in soup.find_all("img"):
            img_url = images.get_img_url(self.url, img)
            if not img_url:
                continue

            image_name = self.download_image(img_url)
            # Ad the articles live in the `content` folder, we need to go one level up
            image_path = join("..", IMAGES, image_name)
            img["src"] = image_path

        return soup.decode()

    def download_image(self, url: str) -> str:
        """
        Download an image from a URL into the images folder
        """
        logger.debug(f"Downloading image {url}...")
        image_name = images.get_image_filename(url)
        image_path = join(self.images_path, image_name)
        images.download_image(url, image_path)
        return image_name

    def get_kwargs(self) -> dict:
        """
        Return a dict of the values necessary for rendering an article
        """
        return dict(title=self.title, author=self.author, date=self.date, content=self.content)


class EPUB:
    """
    Represents the whole EPUB book
    """

    def __init__(self, articles: List[EPUBArticle], title: str, epub_path: str):
        """
        Constructor
        """
        self.articles = articles
        self.title = title
        self.id = normalize_str(self.title)
        self.uuid = uuid4()
        # The date must be in the ISO 8601 format: https://www.w3.org/TR/NOTE-datetime
        self.date = arrow.now().isoformat()
        self._dst_path = epub_path

    def prepare_epub_dirs(self) -> None:
        """
        Create all the basic folders in the EPUB archive
        """
        logger.debug("Creating EPUB folders...")
        makedirs(join(self._dst_path, META_INF))
        makedirs(join(self._dst_path, OEBPS, CONTENT))
        makedirs(join(self._dst_path, OEBPS, IMAGES))

    def build(self) -> str:
        """
        Create an EPUB file from articles and templates

        :return: Path to the created epub archive
        """
        logger.info(f"Creating an EPUB book for `{self.title}`...")
        try:
            self.prepare_epub_dirs()
            self.copy_fixed_files()

            self.render_title()
            self.render_ncx_toc()
            self.render_html_toc()
            self.render_articles()
            self.render_opf()

            epub_path = self.compress_epub()
            logger.info("Successfully created an EPUB archive!")
            return epub_path
        finally:
            rmtree(self._dst_path)

    def compress_epub(self) -> str:
        """
        Create the EPUB ZIP archive

        First add the `mimetype` file - EPUB specs say it must be first and uncompressed.
        Then, recursively add all the files and folders under META-INF and OEBPS
        """
        logger.debug("Creating an epub archive...")

        epub_name = f"{self.id}.epub"
        epub_path = join(mkdtemp(prefix="epub"), epub_name)

        # The EPUB must contain the META-INF and mimetype files at the root, so
        with ZipFile(epub_path, "w") as epub:
            # Add the mimetype file first (as is required by EPUB format) and set it to be uncompressed
            epub.write(join(self._dst_path, MIMETYPE), arcname=MIMETYPE, compress_type=ZIP_STORED)
            self.recursively_add_files_to_epub_archive(epub)
        return epub_path

    def recursively_add_files_to_epub_archive(self, epub: ZipFile) -> None:
        """
        Recursively add all the files in the EPUB dest folder to the epub archive
        """
        for dirname, subdirs, files in walk(self._dst_path):
            # Skip the root folder, as the only file (as opposed to folder) there is the mimefile
            if dirname == self._dst_path:
                continue
            # This will be the file/folder name relative to the archive root
            relative_dirname = dirname.replace(self._dst_path, ".")

            # Add the folder itself to the archive, then add all the files in it
            epub.write(filename=dirname, arcname=relative_dirname)
            self.add_files_to_zip(epub, files, dirname, relative_dirname)

    @staticmethod
    def add_files_to_zip(epub: ZipFile, files: List[str], dirname: str, relative_dirname: str) -> None:
        """
        Recursively add all the files in a list to the epub archive
        """
        for filename in files:
            epub.write(
                filename=join(dirname, filename),
                arcname=join(relative_dirname, filename),
                compress_type=ZIP_DEFLATED,
                compresslevel=9,
            )

    def render_articles(self) -> None:
        """
        Go over all the articles and write their formatted content to disk

        For each article:
            1. Parse its content (also downloads images)
            2. If the parse did not succeed do nothing
            3. If the article was parsed successfully, use the `article.xhtml` template to create the final article
        """
        logger.debug("Rendering articles...")
        parser_cls = self._get_parser_class()
        with parser_cls() as parser:
            for article in self.articles:
                if not article.parse(parser):
                    continue
                kwargs = dict(title=article.title, author=article.author, date=article.date, content=article.content)
                article_path = join(OEBPS, CONTENT, f"{article.id}.xhtml")
                article_html = self.render_template(join(OEBPS, CONTENT, "article.xhtml"), **kwargs)
                self.write_file(article_html, article_path)

    @staticmethod
    def _get_parser_class() -> Type[ParserBase]:
        """Importing the classes here to avoid issues with optional packages (e.g. docker)"""
        parser_type = Parser(config.parser)

        if parser_type == Parser.MERCURY:
            from .mercury_parser import MercuryParser

            return MercuryParser
        elif parser_type == Parser.READABILITY:
            from .readability_parser import ReadabilityParser

            return ReadabilityParser
        else:
            raise ValueError(f"Parser must be one of: {Parser.__values__}")

    def render_opf(self) -> None:
        """
        Render the content.opf XML file

        content.opf requires the following:
            1. <item> elements for all the articles in the <manifest> section
            2. <item> elements for all the images in the <manifest> section
            3. <itemref> elements for all the articles in the <spine> section
        """
        logger.debug("Generating content.opf...")
        manifest_articles = self.generate_manifest_articles()
        manifest_images = self.generate_manifest_images()
        spine_articles = self.generate_spine_articles()
        kwargs = dict(
            title=self.title,
            date=self.date,
            uuid=self.uuid,
            manifest_articles=manifest_articles,
            manifest_images=manifest_images,
            spine_articles=spine_articles,
        )
        self.render_and_write(join(OEBPS, "content.opf"), **kwargs)

    def generate_manifest_images(self) -> str:
        """
        Create <item> elements for all the images in the <manifest> section
        """
        logger.debug("Generating manifest images...")
        manifest_image_template = Template('<item id="${id}" href="images/${id}" media-type="image/${ext}"/>')
        manifest_images: List[dict] = [
            dict(id=image_name, ext=images.get_img_extension(image_name))
            for image_name in listdir(join(self._dst_path, OEBPS, IMAGES))
            if image_name != "cover.png"
        ]
        return "\n\t".join([manifest_image_template.substitute(**image) for image in manifest_images])

    def generate_manifest_articles(self) -> str:
        """
        Create <item> elements for all the articles in the <manifest> section
        """
        logger.debug("Generating manifest articles...")
        manifest_article_template = Template(
            '<item id="${id}" href="content/${id}.xhtml" media-type="application/xhtml+xml"/>'
        )
        return "\n\t".join([manifest_article_template.substitute(id=article.id) for article in self.articles])

    def generate_spine_articles(self) -> str:
        """
        Create <itemref> elements for all the articles in the <spine> section
        """
        logger.debug("Generating spine articles...")
        spine_article_template = Template('<itemref idref="${id}"/>')
        return "\n\t".join([spine_article_template.substitute(id=article.id) for article in self.articles])

    def render_and_write(self, relative_path: str, **kwargs: Any) -> None:
        """
        Render a template and write it to the destination path
        """
        content = self.render_template(relative_path, **kwargs)
        self.write_file(content=content, path=relative_path)

    def render_title(self) -> None:
        """
        Render and write the title.xhtml file (title page of the book)
        """
        logger.debug("Rendering title.xhtml...")
        self.render_and_write(join(OEBPS, "title.xhtml"), **dict(title=self.title))

    def render_ncx_toc(self) -> None:
        """
        Generate, render and write the toc.ncx file (Table of Contents)
        """
        logger.debug("Rendering toc.ncx...")
        navpoints = self.generate_navpoints()
        kwargs = dict(title=self.title, uuid=self.uuid, navpoints=navpoints)
        self.render_and_write(join(OEBPS, "toc.ncx"), **kwargs)

    def render_html_toc(self) -> None:
        """
        Generate, render and write the toc.xhtml file (Table of Contents)
        """
        logger.debug("Rendering toc.xhtml...")
        toc = self.generate_html_toc()
        kwargs = dict(toc=toc, title=self.title)
        self.render_and_write(join(OEBPS, "toc.xhtml"), **kwargs)

    def generate_html_toc(self) -> str:
        """
        Create an HTML <li> element per article
        """
        template = Template('<li><a href="content/${id}.xhtml">${title}</a></li>')
        toc = [dict(id=article.id, title=article.title) for i, article in enumerate(self.articles)]
        return "\n\t".join([template.substitute(**elem) for elem in toc])

    def generate_navpoints(self) -> str:
        """
        Create a navpoint per article for use in the toc.ncx file
        """
        logger.debug("Generating navpoints...")
        template = self.get_template(join(_R2K, "navpoint.xml"))
        navpoints = [
            dict(id=article.id, title=article.title, order=i + NAVPOINT_OFFSET)
            for i, article in enumerate(self.articles)
        ]
        return "\n\t\t".join([template.substitute(**navpoint) for navpoint in navpoints])

    def copy_fixed_files(self) -> None:
        """
        Copy to the destination folder all the files that don't require rendering and can be copied as is
        """
        logger.debug("Copying fixed files...")
        self.copy_file("mimetype")
        self.copy_file(join(META_INF, "container.xml"))
        self.copy_file(join(OEBPS, IMAGES, "cover.png"))
        self.copy_file(join(OEBPS, "stylesheet.css"))

    def render_template(self, template_path: str, **kwargs: Any) -> str:
        """
        Return a rendered template
        """
        template = self.get_template(template_path)
        return template.substitute(**kwargs)

    @staticmethod
    def get_template(template_path: str) -> Template:
        """
        Return a Template object from the templates EPUB dir
        """
        with open(join(EPUB_DIR, template_path)) as f:
            return Template(f.read())

    def write_file(self, content: str, path: str) -> None:
        """
        Write a text file (not a binary one) to the EPUB destination path
        """
        with open(join(self._dst_path, path), "w") as f:
            f.write(content)

    def copy_file(self, template_path: str) -> None:
        """
        Copy a file from the template folder to the destination path
        """
        src_path = join(EPUB_DIR, template_path)
        dst_path = join(self._dst_path, template_path)
        copyfile(src_path, dst_path)



from lxml.html import tostring
import lxml.html
import re

from .cleaners import normalize_spaces, clean_attributes
from .encoding import get_encoding
from .compat import str_

utf8_parser = lxml.html.HTMLParser(encoding="utf-8")


def build_doc(page):
    if isinstance(page, str_):
        encoding = None
        decoded_page = page
    else:
        encoding = get_encoding(page) or "utf-8"
        decoded_page = page.decode(encoding, "replace")

    # XXX: we have to do .decode and .encode even for utf-8 pages to remove bad characters
    doc = lxml.html.document_fromstring(
        decoded_page.encode("utf-8", "replace"), parser=utf8_parser
    )
    return doc, encoding


def js_re(src, pattern, flags, repl):
    return re.compile(pattern, flags).sub(src, repl.replace("$", "\\"))


def normalize_entities(cur_title):
    entities = {
        u"\u2014": "-",
        u"\u2013": "-",
        u"&mdash;": "-",
        u"&ndash;": "-",
        u"\u00A0": " ",
        u"\u00AB": '"',
        u"\u00BB": '"',
        u"&quot;": '"',
    }
    for c, r in entities.items():
        if c in cur_title:
            cur_title = cur_title.replace(c, r)

    return cur_title


def norm_title(title):
    return normalize_entities(normalize_spaces(title))


def get_title(doc):
    title = doc.find(".//title")
    if title is None or title.text is None or len(title.text) == 0:
        return "[no-title]"

    return norm_title(title.text)


def add_match(collection, text, orig):
    text = norm_title(text)
    if len(text.split()) >= 2 and len(text) >= 15:
        if text.replace('"', "") in orig.replace('"', ""):
            collection.add(text)


TITLE_CSS_HEURISTICS = [
    "#title",
    "#head",
    "#heading",
    ".pageTitle",
    ".news_title",
    ".title",
    ".head",
    ".heading",
    ".contentheading",
    ".small_header_red",
]


def shorten_title(doc):
    title = doc.find(".//title")
    if title is None or title.text is None or len(title.text) == 0:
        return ""

    title = orig = norm_title(title.text)

    candidates = set()

    for item in [".//h1", ".//h2", ".//h3"]:
        for e in list(doc.iterfind(item)):
            if e.text:
                add_match(candidates, e.text, orig)
            if e.text_content():
                add_match(candidates, e.text_content(), orig)

    for item in TITLE_CSS_HEURISTICS:
        for e in doc.cssselect(item):
            if e.text:
                add_match(candidates, e.text, orig)
            if e.text_content():
                add_match(candidates, e.text_content(), orig)

    if candidates:
        title = sorted(candidates, key=len)[-1]
    else:
        for delimiter in [" | ", " - ", " :: ", " / "]:
            if delimiter in title:
                parts = orig.split(delimiter)
                if len(parts[0].split()) >= 4:
                    title = parts[0]
                    break
                elif len(parts[-1].split()) >= 4:
                    title = parts[-1]
                    break
        else:
            if ": " in title:
                parts = orig.split(": ")
                if len(parts[-1].split()) >= 4:
                    title = parts[-1]
                else:
                    title = orig.split(": ", 1)[1]

    if not 15 < len(title) < 150:
        return orig

    return title


# is it necessary? Cleaner from LXML is initialized correctly in cleaners.py
def get_body(doc):
    for elem in doc.xpath(".//script | .//link | .//style"):
        elem.drop_tree()
    # tostring() always return utf-8 encoded string
    # FIXME: isn't better to use tounicode?
    raw_html = str_(tostring(doc.body or doc))
    cleaned = clean_attributes(raw_html)
    try:
        # BeautifulSoup(cleaned) #FIXME do we really need to try loading it?
        return cleaned
    except Exception:  # FIXME find the equivalent lxml error
        # logging.error("cleansing broke html content: %s\n---------\n%s" % (raw_html, cleaned))
        return raw_html



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'python-readability'}

packages = \
['r2k',
 'r2k.cli',
 'r2k.cli.config',
 'r2k.cli.feed',
 'r2k.cli.kindle',
 'r2k.ebook',
 'r2k.gmail',
 'readability',
 'readability.compat']

package_data = \
{'': ['*']}

install_requires = \
['arrow>=0.15.8,<0.16.0',
 'beautifulsoup4>=4.9.0,<5.0.0',
 'chardet>=3.0.4,<4.0.0',
 'click>=7.1.1,<8.0.0',
 'cssselect>=1.1.0,<2.0.0',
 'feedparser>=5.2.1,<6.0.0',
 'lxml>=4.5.2,<5.0.0',
 'orjson>=2.6.6,<3.0.0',
 'pick>=0.6.7,<0.7.0',
 'python-dateutil>=2.8.1,<3.0.0',
 'pyyaml>=5.3.1,<6.0.0',
 'requests>=2.24.0,<3.0.0']

extras_require = \
{'docker': ['docker>=4.2.0,<5.0.0']}

entry_points = \
{'console_scripts': ['r2k = r2k.cli:main']}

setup_kwargs = {
    'name': 'r2k',
    'version': '0.7.3',
    'description': 'A tool that lets you periodically send articles received from an RSS feed to your Kindle',
    'long_description': '# R2K (RSS to Kindle)\n\n`r2k` is a tool that lets you to track your favorite RSS/Atom feeds, and send cleaned-up (i.e. only text\nand images) articles received from them directly to your Kindle.\n\n[Installation](#installation)\n\n[Cleaning up the articles](#cleaning-up-the-articles)\n\n[Usage](#usage)\n\n\n## Installation\nThe best way to get up and running with `r2k` us to use pipx. Simply run:\n```bash\npipx install r2k\n```\n\nIf you\'re not using `pipx` (you should!), `pip install r2k` will also do.\n\nIf you want to add `r2k` to your Poetry project, run `poetry add r2k`.\n\n## Cleaning up the articles\n\nThere are currently 3 ways to clean up articles:\n1. The default is using the `readability` Python implementation. It doesn\'t require any extra steps or \ninstallations, and thus is the easiest to use. It is, however, a bit less precise than the other 2 methods. \n1. Using the [PushToKindle](http://pushtokindle.com/) service. The service works by attaching an email \naddress that forwards cleaned up versions of URLs to your Kindle. It\'s free for a certain amount of articles,\nbut you need to become their supporter afterward.\n1. Using a dockerized version of the [Mercury-Parser](https://github.com/postlight/mercury-parser). \nThis method requires Docker, but is marginally better than the `readability` approach.\n \n## Usage\n\n### Preparations\n\n#### If you\'re using PushToKindle (PTK)\n\nThe free version of PTK only allows for 25 articles to be sent using their service. After this,\nyou\'ll have to become their "sustainer" (for as low as 1$/month) on Patreon \n[here](https://www.patreon.com/bePatron?c=1946606).\n\nBefore using `r2k` with PTK you need to:\n\n1. Know your kindle email address (find it [here](https://www.amazon.com/mycd), under \n"Preferences" -> "Personal Document Settings").\n\n2. Add `kindle@fivefilters.org` to the list of approved email senders (in the same place in \nAmazon\'s settings).\n\n### Set up your configuration file\n\nMost of what `r2k` does involves the configuration file, in which the feeds you\'re subscribed to\nare kept, as well as some other data.\n\nAfter installation, run:\n\n```bash\nr2k config init [-p CONFIG_PATH] \n```\n\nThe default location for the config YAML file is in `~/.r2k/config.yml`.\n\nDuring the init you\'ll be asked several questions (like your kindle email address).\n\nTo see your configuration run:\n\n```bash\nr2k config show [-p CONFIG_PATH]\n```\n\n#### Adding your GMail credentials\n\nCurrently `r2k` relies on the fact that you have a GMail address, and it requires a \n[Google App Password](https://support.google.com/accounts/answer/185833?hl=en&authuser=1).\nThese are useful in cases where you don\'t want to go through 2FA authentication. Go\n[here](https://myaccount.google.com/u/1/apppasswords) to generate an `r2k` App Password, and \nadd it to the configuration.\n\n### Add some RSS subscriptions\n\n#### Using an OPML file\n\nThe [OPML](https://en.wikipedia.org/wiki/OPML) format is widely used in the RSS/Atom world \n(as well as in podcasting and other areas) to represent a collection of feeds. You can export your \nexisting subscriptions from most feed readers into an OPML file.\n\nTo load all of your subscriptions in one move run:\n\n```bash\nr2k feed import PATH_TO_OPML_FILE\n```\n\n#### Manually adding feeds\n\nIf you don\'t have an OPML export, or just want to add a single feed you can run:\n\n```bash\nr2k feed add -t FEED_TITLE -u FEED_URL\n```\n\nIf the `FEED_URL` is a proper RSS feed (i.e. an actual XML feed URL), it will be added as is.\nIf the `FEED_URL` is a regular URL, `r2k` will attempt to find the RSS feed by analyzing the page\nsource. In the case of multiple candidates (e.g. WordPress content feed and comment feed), you will\nbe presented with a list of choices.\n\n### Send updates to your Kindle\n\nRight now the "periodical" part of `r2k` is not yet operational. In order to send updates to your\nKindle you\'ll have to run:\n\n```bash\nr2k kindle send [-f FEED_TITLE]\n```\n\nIf you don\'t pass the `-f/--feed-title` option, updates will be sent for all of your subscriptions.\n\nThe first time that `kindle send` is run for any feed, you will be presented with a list of all the \navailable articles in the feed (note that RSS feeds usually only keep a subset of the most recent\nentries), and will be asked to choose the last one you\'ve already read. This is to avoiding sending\nyou any article you\'ve already consumed.\n ',
    'author': 'Pavel Brodsky',
    'author_email': 'mcouthon@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mcouthon/r2k',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import click

from .config_init import config_init
from .config_set import config_set
from .config_show import config_show


@click.group()
def config() -> None:
    """Interact with the application config."""
    pass


config.add_command(config_init)
config.add_command(config_show)
config.add_command(config_set)



import click

from r2k.cli import logger
from r2k.constants import Parser


class Prompt:
    """A convenience class to get the correct prompt per config value"""

    send_from = dict(text="Please provide your gmail email address")
    password = dict(text="Please provide your gmail app password", hide_input=True)
    kindle_address = dict(text="Please provide your free kindle address (e.g. my_kindle@kindle.com)")
    parser = dict(
        text="Please choose the parser you're going to use",
        default=Parser.READABILITY.value,
        type=click.Choice(Parser.__values__),
    )

    @classmethod
    def get(cls, key: str) -> str:
        """Get a config value (using click's prompt)"""
        kwargs = getattr(cls, key, dict(text=f"Please provide a new value for {key}"))
        return logger.prompt(**kwargs)



from datetime import datetime
from typing import Union

import arrow
import yaml

DATE_FORMATS = [arrow.FORMAT_RSS, arrow.FORMAT_ATOM, "ddd, DD MMM YYYY HH:mm:ss ZZZ"]


def get_pretty_date_str(
    date: Union[str, datetime, arrow.Arrow], show_time: bool = False, show_year: bool = True
) -> str:
    """Return a nice representation of a date"""
    date = parse_date(date)

    # Only day and Month
    template = "%d %B"

    if show_year:
        template = f"{template}, %Y"

    # Add the time
    if show_time:
        template = f"%H:%M {template}"

    return date.strftime(template)


def parse_date(date: Union[str, datetime, arrow.Arrow]) -> arrow.Arrow:
    """Return a standardized UTC Arrow object"""
    _date = None
    if isinstance(date, str):
        try:
            _date = arrow.get(date)
        except arrow.ParserError:
            pass

        if not _date:
            parser = arrow.parser.DateTimeParser()
            for fmt in DATE_FORMATS:
                try:
                    _date = parser.parse(date, fmt=fmt)
                except arrow.ParserError:
                    continue
        if not _date:
            raise arrow.ParserError()

    elif isinstance(date, datetime):
        _date = arrow.Arrow.fromdatetime(date)
    else:
        _date = date

    return _date.astimezone(tz=arrow.utcnow().tzinfo)


def now() -> arrow.Arrow:
    """Convenience function for getting the current UTC DateTime"""
    return arrow.utcnow()


def arrow_representer(dumper: yaml.Dumper, data: arrow.Arrow) -> str:
    """
    Represent an `arrow.arrow.Arrow` object as a scalar in ISO format.

    ! '2013-05-07T04:24:24+00:00'
    """
    return dumper.represent_scalar("!", data.isoformat("T"))


yaml.add_representer(arrow.Arrow, arrow_representer)
yaml.SafeDumper.add_representer(arrow.Arrow, arrow_representer)
